import { rabbitMq, ums, mail, SMTP_USER, DEBUG_EMAIL } from '../variables';
import { log } from '../logger';
import { wsReturnKey, type AppearanceMap, type InsightModel } from 'ts-common';
import { createHeadlessEditor } from '@lexical/headless';
import { $generateHtmlFromNodes } from '@lexical/html';
import { JSDOM } from 'jsdom';
import { Query2BackendQuery } from '../utils/reactflow/query2backend';
import { query2Cypher } from '../utils/cypher/converter';
import { queryService } from './queryService';
import { statCheck } from './statCheck';
import { diffCheck } from './diffCheck';
import { appearanceCheck, sortAndThreshold } from './appearanceCheck';
import { VariableNode } from '../utils/lexical';
import { populateTemplate } from '../utils/insights';
import { RabbitMqBroker } from 'ts-common/rabbitMq';

const dom = new JSDOM();
function setUpDom() {
  const _window = global.window;
  const _document = global.document;
  const _documentFragment = global.DocumentFragment;
  const _navigator = global.navigator;

  // @ts-expect-error - global.window is readonly
  global.window = dom.window;
  global.document = dom.window.document;
  global.DocumentFragment = dom.window.DocumentFragment;
  global.navigator = dom.window.navigator;

  return () => {
    global.window = _window;
    global.document = _document;
    global.DocumentFragment = _documentFragment;
    global.navigator = _navigator;
  };
}

export const insightProcessor = async (publisher: RabbitMqBroker) => {
  /*
  if (mail == null) {
    log.warn('Mail is not configured. Insight processor will be disabled');
    return;
  }
  */
  log.info('Starting insight processor');

  const insightProcessorConsumer = await new RabbitMqBroker(
    rabbitMq,
    'insight-processor',
    `insight-processor`,
    `insight-processor`,
  ).connect();
  log.info('Connected to RabbitMQ ST!');

  async function sendMail(name: string, html: string, recipient: string) {
    if (mail == null) {
      log.warn('Mail is not configured. Insight processor will be disabled');
      return;
    }

    if (DEBUG_EMAIL) {
      log.warn('DEBUG: Would have sent mail to', recipient);
    }

    log.debug('Sending mail to', recipient);
    await mail.sendMail({
      to: recipient,
      from: SMTP_USER,
      subject: `GraphPolaris report: ${name}`,
      html: html,
    });
    log.info('Mail sent to ', recipient);
  }

  await insightProcessorConsumer.startConsuming<{ insight: InsightModel; force: boolean; appearances?: boolean }>(
    'query-service',
    async (message, headers) => {
      let insight = message.insight;
      if (insight == null || insight.template == null || insight.userId == null || insight.saveStateId == null) {
        log.error('Invalid Insight received in insightProcessorConsumer:', insight);
        return;
      }

      if (insight.alarmMode === 'disabled' && !message.force) {
        log.info('Alarm mode is disabled', insight.id);
        return;
      }

      if (insight.recipients == null || insight.recipients.length === 0) {
        log.info('No recipients found in the insight, skipping');
        return;
      }

      if (insight.userId == null) return;
      const ss = await ums.getUserSaveState(insight.userId, insight.saveStateId);
      // log.warn('ss', ss);
      log.warn('insight', insight);

      if (message.appearances && insight.alarmMode === 'entityAppearances') {
        log.info('Received insight appearance request', insight);
        const queries = ss.queryStates.openQueryArray;
        const nodeAppearances: AppearanceMap = [];

        for (const queryIndex in queries) {
          const visualQuery = ss.queryStates.openQueryArray[queryIndex].graph;
          const queryBuilderSettings = ss.queryStates.openQueryArray[queryIndex].settings;
          const convertedQuery = Query2BackendQuery(ss.id, visualQuery, queryBuilderSettings, []);
          const cypher = query2Cypher(convertedQuery);
          if (cypher == null) return;
          try {
            const result = await queryService(ss.dbConnections[0], cypher, true);
            appearanceCheck(result, insight, ss, Number(queryIndex), nodeAppearances);
          } catch (err) {
            log.error('Error processing insight appearances', err);
          }
        }

        publisher.publishMessageToFrontend(
          {
            type: wsReturnKey.insightAppearance,
            callID: headers.callID,
            value: nodeAppearances,
            status: 'success',
          },
          headers.routingKey,
          headers,
        );
        return;
      }

      log.info('Received insight to be processed', insight);

      const editor = createHeadlessEditor({
        nodes: [VariableNode],
        onError: error => {
          log.error(error);
        },
      });

      editor.update(() => {
        const state = editor.parseEditorState(JSON.parse(insight.template));
        editor.setEditorState(state);
      });

      const queries = ss.queryStates.openQueryArray;
      const visualizations = ss.visualizations.openVisualizationArray;

      const nodeAppearances: AppearanceMap = [];

      for (const queryIndex in queries) {
        const visualQuery = ss.queryStates.openQueryArray[queryIndex].graph;
        const queryBuilderSettings = ss.queryStates.openQueryArray[queryIndex].settings;
        const convertedQuery = Query2BackendQuery(ss.id, visualQuery, queryBuilderSettings, []);
        const cypher = query2Cypher(convertedQuery);
        if (cypher == null) return;
        try {
          const result = await queryService(ss.dbConnections[0], cypher, true);

          insight.status = false;

          if (insight.alarmMode === 'always') {
            insight.status = true;
          } else if (insight.alarmMode === 'diff') {
            insight = await diffCheck(insight, ss, result);
          } else if (insight.alarmMode === 'conditional' && insight.conditionsCheck && insight.conditionsCheck.length > 0) {
            insight = statCheck(insight, result);
          } else if (insight.alarmMode === 'entityAppearances') {
            appearanceCheck(result, insight, ss, Number(queryIndex), nodeAppearances);
            continue; // mail is sent after the for loop, for all queries at once
          }

          if (insight.userId == null) return; // fixes ts but never is the case
          await ums.updateInsight(insight.userId, insight.id, insight);

          if (insight.status || message.force) {
            if (insight.status) log.debug('Insight passed the check');
            if (message.force) log.debug('Forced insight processing');

            editor.read(async () => {
              const cleanUpDom = setUpDom();
              let html = $generateHtmlFromNodes(editor);
              cleanUpDom();

              html = await populateTemplate(html, result, visualizations);

              for (const recipient of insight.recipients) {
                sendMail(insight.name, html, recipient);
              }
            });
          } else {
            log.debug('WARN: Insight did not pass the check');
          }
        } catch (err) {
          log.error('Error processing insight', err);
        }
      }

      if (insight.alarmMode === 'entityAppearances') {
        const nodeAppearancesSortedTH = sortAndThreshold(nodeAppearances, insight);
        log.info('nodeAppearancesSortedTH ', nodeAppearancesSortedTH);
        // Sent a single report for all queries at once
        editor.read(async () => {
          const cleanUpDom = setUpDom();
          let html = $generateHtmlFromNodes(editor);
          cleanUpDom();

          html = await populateTemplate(html, undefined, visualizations, nodeAppearancesSortedTH);
          for (const recipient of insight.recipients) {
            sendMail(insight.name, html, recipient);
          }
        });
      }
    },
  );
};
